@@ -1,6 +1,7 @@
 """
 The freshest in-order generator parser
 """
+
 from __future__ import annotations
 
 import ast as pyast
@@ -206,11 +207,14 @@ class Function:
             raise UnsupportedSyntaxError.from_pyast(ret, self.__context.name) from e
 
         self.__context.validate()
-        head, tail = self._weave_nonclocked_edges(
-            self._create_assign_nodes(self.__context.output_vars, stmts, prefix),
-            f"{prefix}_e",
-        )
-        tail.edge = ir.ClockedEdge(unique_id=f"{prefix}_last_e")
+        try:
+            head, tail = self._weave_nonclocked_edges(
+                self._create_assign_nodes(self.__context.output_vars, stmts, prefix),
+                f"{prefix}_e",
+            )
+            tail.edge = ir.ClockedEdge(unique_id=f"{prefix}_last_e")
+        except Exception as e:
+            raise RuntimeError(f"{e} in {self.__context.name}") from e
         return head, [tail.edge]
 
     def _parse_stmts(
@@ -910,26 +914,28 @@ class Function:
 
         With special case for floor division
         """
+        left = self._parse_expression(expr.left)
+        right = self._parse_expression(expr.right)
+
         if isinstance(expr.op, pyast.Add):
-            return ir.Add(
-                self._parse_expression(expr.left), self._parse_expression(expr.right)
-            )
+            return ir.Add(left, right)
         if isinstance(expr.op, pyast.Sub):
-            return ir.Sub(
-                self._parse_expression(expr.left), self._parse_expression(expr.right)
-            )
-
+            return ir.Sub(left, right)
         if isinstance(expr.op, pyast.Mult):
-            return ir.Mul(
-                self._parse_expression(expr.left), self._parse_expression(expr.right)
-            )
-
+            return ir.Mul(left, right)
         if isinstance(expr.op, pyast.FloorDiv):
-            left = self._parse_expression(expr.left)
-            right = self._parse_expression(expr.right)
             return ir.FloorDiv(left, right)
         if isinstance(expr.op, pyast.Mod):
-            left = self._parse_expression(expr.left)
-            right = self._parse_expression(expr.right)
             return ir.Mod(left, right)
+        if isinstance(expr.op, pyast.BitOr):
+            return ir.BitOr(left, right)
+        if isinstance(expr.op, pyast.BitAnd):
+            return ir.BitAnd(left, right)
+        if isinstance(expr.op, pyast.BitXor):
+            return ir.BitXor(left, right)
+        if isinstance(expr.op, pyast.LShift):
+            return ir.LShift(left, right)
+        if isinstance(expr.op, pyast.RShift):
+            return ir.RShift(left, right)
+
         raise UnsupportedSyntaxError(f"Unsupported binop `{pyast.dump(expr.op)}")
