@@ -18,6 +18,11 @@ from .expressions import (
     Mod,
     Mul,
     Pow,
+    BitAnd,
+    BitOr,
+    BitXor,
+    LShift,
+    RShift,
     State,
     Sub,
     Ternary,
@@ -32,7 +37,6 @@ from .graph import (
     BasicElement,
     BasicNode,
     ClockedEdge,
-    DoneNode,
     Edge,
     Element,
     IfElseNode,
@@ -41,3 +45,4 @@ from .graph import (
     create_cytoscape_elements,
     create_networkx_adjacency_list,
 )
+from .instance import Instance
