@@ -7,7 +7,6 @@ from __future__ import annotations
 
 import ast
 import copy
-import io
 import logging
 from dataclasses import dataclass, field
 from itertools import zip_longest
@@ -31,6 +30,8 @@ class Context(GenericReprAndStr):
     """
 
     # pylint: disable=too-many-instance-attributes, too-many-public-methods
+    _frozen: bool = False  # Make immutable
+
     name: str = ""
     testbench_suffix: str = "_tb"
     is_generator: bool = False
@@ -68,6 +69,17 @@ class Context(GenericReprAndStr):
         default_factory=dict
     )  # generator instances
 
+    def freeze(self):
+        """
+        Freeze this context to be immutable
+        """
+        self._frozen = True
+
+    def __setattr__(self, attr, value):
+        if self._frozen:
+            raise AttributeError("Frozen")
+        return super().__setattr__(attr, value)
+
     @classmethod
     def empty_valid(cls):
         """
@@ -87,14 +99,6 @@ class Context(GenericReprAndStr):
         """
         return f"{self.name}{self.testbench_suffix}"
 
-    def _repr(self):
-        """
-        Avoids recursion on itself
-        """
-        dic = copy.deepcopy(self.__dict__)
-        del dic["namespace"]
-        return dic
-
     def _use_input_type_hints(self):
         """
         Uses input type hints
@@ -116,7 +120,7 @@ class Context(GenericReprAndStr):
         try:
             self.input_types = list(map(input_mapper, input_args))
         except Exception as e:
-            raise TypeInferenceError() from e
+            raise TypeInferenceError(self.name) from e
 
     def _use_output_type_hints(self):
         """
@@ -146,7 +150,7 @@ class Context(GenericReprAndStr):
         try:
             self.output_types = list(map(output_mapper, output_args))
         except Exception as e:
-            raise TypeInferenceError(f"in function `{self.name}`") from e
+            raise TypeInferenceError(self.name) from e
         self.default_output_vars()
 
     def validate(self):
@@ -189,46 +193,20 @@ class Context(GenericReprAndStr):
         Python ast node rooted at function
         """
         assert isinstance(self._py_ast, ast.FunctionDef)
-        return self._py_ast
+        return copy.deepcopy(self._py_ast)
 
     @py_ast.setter
     def py_ast(self, other: ast.FunctionDef):
         assert isinstance(other, ast.FunctionDef)
         self._py_ast = other
 
-    @property
-    def testbench_file(self):
-        """
-        Testbench stream
-        """
-        assert isinstance(self._testbench_file, io.IOBase)
-        return self._testbench_file
-
-    @testbench_file.setter
-    def testbench_file(self, other: io.IOBase):
-        assert isinstance(other, io.IOBase)
-        self._testbench_file = other
-
-    @property
-    def module_file(self):
-        """
-        Module stream
-        """
-        assert isinstance(self._module_file, io.IOBase), type(self._module_file)
-        return self._module_file
-
-    @module_file.setter
-    def module_file(self, other: io.IOBase):
-        assert isinstance(other, io.IOBase)
-        self._module_file = other
-
     @property
     def entry_state(self):
         """
         The first state that does work in the graph representation
         """
         assert isinstance(self._entry_state, State), self
-        return self._entry_state
+        return copy.deepcopy(self._entry_state)
 
     @entry_state.setter
     def entry_state(self, other: State):
@@ -241,7 +219,7 @@ class Context(GenericReprAndStr):
         The ready state
         """
         assert isinstance(self._done_state, State), self
-        return self._done_state
+        return copy.deepcopy(self._done_state)
 
     @done_state.setter
     def done_state(self, other: State):
@@ -249,7 +227,7 @@ class Context(GenericReprAndStr):
         self._done_state = other
 
     @property
-    def input_vars(self):
+    def input_vars(self) -> list[Var]:
         """
         Input variables
         """
@@ -261,37 +239,38 @@ class Context(GenericReprAndStr):
         self._input_vars = typed_list(other, Var)
 
     @property
-    def output_vars(self):
+    def output_vars(self) -> list[Var]:
         """
         Output variables
         """
         assert guard(self._output_vars, list), f"Unknown output variables {self}"
-        return self._output_vars
+        return copy.deepcopy(self._output_vars)
 
     def default_output_vars(self):
         """
         Sets own output vars to default based on number of output variables
         """
-        assert self.output_types is not None
+        assert self.output_types is not None, f"{self.name}"
         self._output_vars = [
             ExclusiveVar(f"{self.prefix}_output_{i}")
             for i in range(len(self.output_types))
         ]
 
-    def refresh_input_vars(self):
+    def refresh_input_output_vars(self):
         """
         Update input vars with prefix
         """
         self._input_vars = list(
             map(lambda x: self.make_var(x.py_name), self.input_vars),
         )
+        self.default_output_vars()
 
     @property
     def local_vars(self) -> list[Var]:
         """
         Gets local variables
         """
-        return self._local_vars
+        return copy.deepcopy(self._local_vars)
 
     def add_local_var(self, var: Var):
         """
@@ -317,7 +296,8 @@ class Context(GenericReprAndStr):
         """
         if var.py_name in self.generator_instances:
             raise StaticTypingError(
-                f"{var.py_name} changed type from generator instance to another type"
+                f"Variable `{var.py_name}` changed type from generator"
+                f" instance to another type in {self.name}"
             )
         if var not in (*self._local_vars, *self.input_vars, *self.output_vars):
             self._local_vars.append(typed_strict(var, Var))
@@ -344,7 +324,7 @@ class Context(GenericReprAndStr):
         assert isinstance(name, str)
         self._states.add(name)
 
-    def __check_types(
+    def _check_types(
         self, expected_types: list[type[Any]], actual_values: list[type[Any]]
     ):
         for expected, actual in zip_longest(
@@ -360,14 +340,14 @@ class Context(GenericReprAndStr):
         Checks if input to functions' types matches previous inputs
         """
         assert guard(self.input_types, list)
-        self.__check_types(self.input_types, input_)
+        self._check_types(self.input_types, input_)
 
     def check_output_types(self, output):
         """
         Checks if outputs to functions' types matches previous outputs
         """
         assert guard(self.output_types, list)
-        self.__check_types(self.output_types, output)
+        self._check_types(self.output_types, output)
 
     def create_generator_instance(self, name: str) -> Instance:
         """
@@ -388,7 +368,7 @@ class Context(GenericReprAndStr):
             )
         )
 
-        signals = ProtocolSignals(prefix=f"{self.prefix}{name}_{self.name}__")
+        signals = ProtocolSignals(prefix=f"{self.prefix}{self.name}_{name}__")
 
         return Instance(
             self.name,
