@@ -23,6 +23,8 @@ def decorator_with_args(func):
             # actual decorated function
             return func(args[0])
         # decorator arguments
-        return lambda realf: func(realf, *args, **kwargs)
+        return lambda real_func: func(real_func, *args, **kwargs)
+
+    new_dec.__doc__ = func.__doc__
 
     return new_dec
